<?php

/**
 * @file
 * Drush commands for MongoDB Indexer.
 */

/**
 * Implements hook_drush_command().
 */
function mongodb_indexer_drush_command() {
  $items = array();
  $items['mongodb-indexer-mark-reindexing'] = array(
    'callback' => 'mongodb_indexer_mark_reindex',
    'description' => 'Mark entity bundle for reindexing.',
    'aliases' => array('mi-mr'),
  );
  $items['mongodb-indexer-start-reindexing'] = array(
    'callback' => 'mongodb_indexer_start_reindexing',
    'description' => 'Index all queue items from reindexing queue.',
    'aliases' => array('mi-i'),
  );

  return $items;
}

/**
 * Implements hook_drush_help().
 */
function mongodb_indexer_drush_help($section) {
  switch ($section) {
    case 'mongodb-indexer-mark-reindexing':
      return dt("Mark entity bundle for reindexing.");

    case 'mongodb-indexer-start-reindexing':
      return dt("Index all queue items from reindexing queue.");
  }
}

/**
 * Mark all selected entity type and bundle items as need to be indexed.
 */
function mongodb_indexer_mark_reindex() {
  // Get selected entity and bundle.
  $engine = \Drupal::service('mongodb_indexer.engine');
  $selected_entity_bundle = $engine->getSelectedEntityBundle();

  $options = array(
    1 => 'all',
  );

  // The next choice should be 2 as 0 is for cancel and 1 is for 'all'.
  $choice_index = 2;

  foreach ($selected_entity_bundle as $key => $row) {
    $bundle = $row['bundle'] ? $row['bundle'] : FALSE;
    if (!$bundle) {
      $label = $row['entity_type'];
    }
    else {
      $label = $row['bundle'] . ' of entity type ' . $row['entity_type'];
    }
    $options[$choice_index] = $label;
    $selected_entity_bundle[$key]['choice'] = $choice_index;
    $selected_entity_bundle[$key]['label'] = $label;
    $choice_index++;
  }

  $choice = drush_choice($options, t('Select entity bundle to reindex'));

  if ($choice == 0) {
    return FALSE;
  }
  foreach ($selected_entity_bundle as $row) {
    // An array contain max MONGODB_INDEXER_DEFAULT_BATCH_SIZE items.
    $items_to_index = array();

    $entity_type = $row['entity_type'];
    $bundle = $row['bundle'];
    $label = $row['label'];

    if (($row['choice'] == $choice) || ($choice == 1)) {
      // Do an EFQ to get all ids of this entity type and bundle.

      $query = \Drupal::entityQuery($entity_type)
        ->condition('status', 1);

      if ($bundle) {
        $query->condition('type', $bundle);
      }

      $entities = $query->execute();

      if (!empty($entities)) {
        $count = 0;
        $total_left = count($entities);
        $output = \Drupal::translation()->formatPlural(
          $total_left,
          "Preparing to add to reindex queue 1 item of {$label}.",
          "Preparing to add to reindex queue @count items of {$label}."
        );

        drush_log($output, "ok");

        foreach ($entities as $id => $entity) {
          // User has a weird default user with id = 0
          if (!$id) {
            continue;
          }
          $items_to_index[] = array(
            'entity_id' => $id,
            'entity_type' => $entity_type,
            'bundle' => $bundle,
          );
          $count++;

          if ($count == \Drupal\mongodb_indexer\IndexerEngineInterface::MONGODB_INDEXER_DEFAULT_BATCH_SIZE) {
            $count = 0;
            if (!empty($items_to_index)) {
              $total_left = $total_left - count($items_to_index);
              $output = \Drupal::translation()->formatPlural(
                count($items_to_index),
                "Adding a batch of 1 item of {$label}. Remaining items {$total_left}.",
                "Adding a batch of @count items of {$label}. Remaining items {$total_left}."
              );

              drush_log($output, "ok");
              $engine->markItemsReindex($items_to_index);

              // Reset item to index.
              $items_to_index = array();
            }
          }
        }

        // This is for the last bit of the items,
        // which could not be exact IndexerEngineInterface::MONGODB_INDEXER_DEFAULT_BATCH_SIZE number.
        if (!empty($items_to_index)) {
          $total_left = $total_left - count($items_to_index);
          $output = \Drupal::translation()->formatPlural(
            count($items_to_index),
            "Adding a batch of 1 item of {$label}. Remaining items {$total_left}.",
            "Adding a batch of @count items of {$label}. Remaining items {$total_left}."
          );

          drush_log($output, "ok");

          $engine->markItemsReindex($items_to_index);

          // Reset item to index.
          $items_to_index = array();
          drush_log("Finised adding items of " . $label . " to queue for reindexing.", "ok");
        }
      }
      else {
        drush_log("No items of " . $label . " found. Nothing was added to the queue.", "ok");
      }
    }
  }

  drush_log("Everything is Done!", "completed");
}

/**
 * Index all queue in queue table.
 */
function mongodb_indexer_start_reindexing() {
  if (!drush_confirm("Are you sure?")) {
    return FALSE;
  }
  drush_log("Start reindexing. This may take a while...", "ok");
  // Create a batch process for processing items in the reindexing queue.
  _mongodb_indexer_batch_reindexing_create();
  // Start the batch process.
  drush_backend_batch_process();

  drush_log("Reindexing done!", "completed");
}

/**
 * Creates and sets a batch for indexing items to mongodb.
 */
function _mongodb_indexer_batch_reindexing_create() {
  $batch = array(
    'title' => t('Reindexing items to MongoDB'),
    'operations' => array(
      array('_mongodb_indexer_process_reindexing_batch', array()),
    ),
    'progress_message' => t('Completed: @current. Remaining approximately: @remaining.'),
    'finished' => '_mongodb_indexer_reindexing_batch_finished',
    'file' => drupal_get_path('module', 'mongodb_indexer') . '/mongodb_indexer.drush.inc',
  );

  batch_set($batch);
}

/**
 * Batch API finishing callback for the MongoDB reindexing functionality.
 *
 * @param bool $success
 *   Result of the batch operation.
 * @param array $results
 *   Results stored from the batch operations.
 * @param array $operations
 *   Remaining batch operation to process.
 */
function _mongodb_indexer_reindexing_batch_finished($success, array $results, array $operations) {
  if ($success) {
    if (!empty($results['reindexed'])) {
      drush_log("Successfully reindexed " . $results['reindexed'] . " items to mongodb.", "ok");
    }
    else {
      drush_log("Could not reindex any items to MongoDB.", "warning");
    }
  }
  else {
    drush_log('An error occurred while reindexing items to MongoDB.', "error");
  }
}

/**
 * Batch API callback for reindexing items to mongodb.
 *
 * @param array $context
 *   An array that will contain information about the
 *   status of the batch. The values in $context will retain their
 *   values as the batch progresses.
 */
function _mongodb_indexer_process_reindexing_batch(array &$context) {
  // Get the number of items remaining in the queue.

  // Get the queue.
  $queueFactory = \Drupal::service('queue');
  $queue = $queueFactory->get("mongodb_indexer_reindex_items");
  $queued_items_count = $queue->numberOfItems();

  if (!isset($context['sandbox']['current'])) {
    // This is the first time the batch operation is run.
    $context['sandbox']['current'] = 1;
    // Set the approximate number of total queued items to be processed.
    $context['sandbox']['total'] = $queued_items_count;
    // Set the approximate number of remaining items to be processed.
    // Count the current one assuming it will be processed successfully.
    $context['sandbox']['remaining'] = $queued_items_count - 1;
    // Store the total number of reindexed items.
    $context['results']['reindexed'] = 0;
  }
  else {
    // This is not the first time the batch operation is run.
    $context['sandbox']['current']++;
    // Set the new approximate number of remaining items.
    // Count the current one assuming it will be processed successfully.
    $context['sandbox']['remaining'] = $queued_items_count - 1;
  }

  // If the current batch process number is >= the total number of
  // queued items, perhaps the initial calculation is no longer valid.
  // It could be that new items have been added to the queue since
  // the batch was first started. Alternatively, this could be indeed
  // the last item to be processed but this cannot be guaranteed.
  // In these situations, increase the total number of items
  // to be more than the current item number.
  if ($context['sandbox']['current'] >= $context['sandbox']['total']) {
    $context['sandbox']['total'] = $context['sandbox']['current'] + 1;
  }

  // Claim a single item (containing an array of entities to be reindexed)
  // from the queue and reindex the entities.
  $engine = \Drupal::service('mongodb_indexer.engine');
  $reindex_result= $engine->indexQueuedReindexItems();

  // If there were no more items that could be claimed from the queue,
  // then the reindexing batch should stop.
  if ($reindex_result === FALSE) {
    $context['finished'] = 1;
  }
  else {
    // Increase the total number of reindexed items.
    $context['results']['reindexed'] += $reindex_result;

    $batch_message = '';
    if ($reindex_result == \Drupal\mongodb_indexer\IndexerEngineInterface::MONGODB_INDEXER_DEFAULT_BATCH_SIZE) {
      // The number of successfully indexed items is the maximum allowed one.
      $batch_message = "Successfully reindexed $reindex_result items.";
    }
    else {
      $batch_message = "Reindexed $reindex_result items, expected ";
      $batch_message .= \Drupal\mongodb_indexer\IndexerEngineInterface::MONGODB_INDEXER_DEFAULT_BATCH_SIZE . ". ";
      $batch_message .= "This is normal if this is the last batch of items to be reindexed.";
    }

    $context['message'] = $batch_message;

    // The operation needs to be repeated. There are probably more items
    // to be claimed from the queue. Provide an estimation about the progress.
    $context['finished'] = $context['sandbox']['current'] / $context['sandbox']['total'];
  }
}

