<?php
namespace Drupal\mongodb_indexer\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Utility\UrlHelper;

class SettingsForm extends FormBase {


  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mongodb_indexer_admin_settings';
  }


  /**
   * @return mixed
   */
  public function getEditableConfigNames() {
    return parent::getEditableConfigNames();
  }


  /**
   * Build the form for the configuration page
   *
   * @param array $form
   * @param FormStateInterface $form_state
   * @return array
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('mongodb_indexer.settings');
    $engine = \Drupal::service('mongodb_indexer.engine');


    $form = array();
    // Check if settings are fine.
    $mongo_client = $engine->getMongoClient();

    if (!$mongo_client) {
      drupal_set_message(t('Unable to connect to MongoDB. Please check your settings.'), 'error');
    }
    else {
      drupal_set_message(t('Connection to MongoDB is working successfully.'));
    }

    // Output queue status.
    if ($mongo_client) {
      $items = $engine->getQueueLength();

      if ($items) {
        drupal_set_message(t('Number of items to be indexed: @items.', array('@items' => $items)), 'warning');
      }
      else {
        drupal_set_message(t('No items in the indexing queue.'));
      }
    }

    // Mongodb settings.
    $form['mongodb_settings'] = array(
      '#type' => 'fieldset',
      '#title' => 'Mongodb settings',
      '#collapsible' => TRUE,
    );

    $form['mongodb_settings']['mongodb_indexer_index_immediately'] = array(
      '#type' => 'checkbox',
      '#title' => t('Index items immediately'),
      '#description' => t('Index items immediately or wait until next cron run. It might cause issues for large sites when this is checked.'),
      '#default_value' => $config->get('mongodb_indexer_index_immediately'),
    );
    $form['mongodb_settings']['mongodb_indexer_items_per_cron_run'] = array(
      '#type' => 'select',
      '#title' => 'Number of items to index per cron run',
      '#description' => t('Higher number could make index operation faster but could also cause timeout and memory limit error.'),
      '#options' => array(0 => 0, 10 => 10, 20=>20, 50=>50, 100=>100, 200=>200, 500=>500, 1000=>1000),
      '#default_value' => $config->get('mongodb_indexer_items_per_cron_run' ),
    );

    // Entity selector to select which entity should be indexed.
    $form['selected_entity'] = array(
      '#type' => 'fieldset',
      '#title' => t('Select contents to index'),
      '#description' => t('Different entity types are saved into separated collections.'),
      '#collapsible' => TRUE,
    );

    $options = $this->getEntityOptions();
    $default_values = $config->get('mongodb_indexer_selected_entity_bundle');

    $form['selected_entity']['mongodb_indexer_selected_entity_bundle'] = array(
      '#type' => 'checkboxes',
      '#options' => $options,
      '#default_value' => $default_values,
    );

    $form['submit'] = array(
      '#type' => 'submit',
      '#value' => t('Submit'),
    );
    return $form;
  }


  /**
   * Get the entity options for the configuration page.
   *
   * @return array
   */
  function getEntityOptions() {
    $entity_info = \Drupal::entityTypeManager()->getDefinitions();
    $engine = \Drupal::service('mongodb_indexer.engine');
    $return_array = array();

    // Loop through all the entity info
    foreach ($entity_info as $key => $val) {

      // If the key is one of the values declared in engines no bundle support array,
      // just use $key -> $value in form.
      if (in_array($key, $engine->getEntityWithoutBundleSupport())) {
        $return_array[$key] = (String)$val->getLabel();
        continue;
      }


      // If however the key is not part of no-bundle-support-array
      // Load the bundles for this entity
      $bundles =  \Drupal::entityManager()->getBundleInfo($key);


      // And loop those through also. This way we can get more fine grained options
      // of what we want to index.
      foreach($bundles as $bundle_key => $bundle) {
        $return_key = $key . '*' . $bundle_key;
        $return_value  = (String)$bundle['label'] . ' of entity type ' . (String)$val->getLabel();
        $return_array[$return_key] = $return_value;
      }
    }


    // Sort the array by its key.
    ksort($return_array);


    return $return_array;
  }


  /**
   * Submit form handler.
   * Save the selected values into configuration.
   *
   * @param array $form
   * @param FormStateInterface $form_state
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    \Drupal::configFactory()->getEditable('mongodb_indexer.settings')
      ->set('mongodb_indexer_selected_entity_bundle', $form_state->getValue('mongodb_indexer_selected_entity_bundle'))
      ->save();

    \Drupal::configFactory()->getEditable('mongodb_indexer.settings')
      ->set('mongodb_indexer_items_per_cron_run', $form_state->getValue('mongodb_indexer_items_per_cron_run'))
      ->save();

    \Drupal::configFactory()->getEditable('mongodb_indexer.settings')
      ->set('mongodb_indexer_index_immediately', $form_state->getValue('mongodb_indexer_index_immediately'))
      ->save();

  }
}
