<?php
namespace Drupal\mongodb_indexer;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database;
use Drupal\mongodb_indexer\Event\IndexerEvent;
use MongoDB\Client;
use MongoDB\BSON\UTCDateTime;

class IndexerEngine implements IndexerEngineInterface {

  protected $mongoClient;
  protected $configFactory;
  protected $collection;
  protected $config;

  protected $queueFactory;


  /**
   * IndexerEngine constructor.
   * @param ConfigFactoryInterface $config_factory
   */
  public function __construct(ConfigFactoryInterface $config_factory) {
    $this->configFactory = $config_factory;
    $this->queueFactory = \Drupal::service('queue');

    $this->config = \Drupal::config('mongodb_indexer.settings');


    // Upon construct, try to create MongoDB connection.
    $this->connect();
  }


  /**
   * Get the MongoDB client object
   * The will try connecting to MongoDB if connection doesn't already exist
   *
   * @return mixed
   */
  function getMongoClient() {
    if(empty($this->mongoClient)) {
      $this->connect();
    }

    return $this->mongoClient;
  }


  /**
   * Get static mongodb client object.
   *
   * @return object
   *   The mongo client object or false if can't create the object
   */
  public function connect() {

    // Load up the settings for the MongoDB
    $mongodb_settings = $this->config->get('driver_settings');


    // If any of the required parameters cannot be found, bail out.
    if (empty($mongodb_settings) || empty($mongodb_settings['hosts'])) {
      return FALSE;
    }


    // Generate the connection string. This supports multiple MongoDB hosts
    $connection_string = 'mongodb://' . implode(',', $mongodb_settings['hosts']);

    $options = !empty($mongodb_settings['options']) ? $mongodb_settings['options'] : array();

    // Make the connection.
    // If connection fails with exception, log it to Drupals default logger
    try {
      $this->mongoClient = new Client($connection_string, $options);
      return TRUE;
    }
    catch (Exception $e) {
      \Drupal::logger('mongodb_indexer')->error($e->getMessage());
      return FALSE;
    }
  }


  /**
   * A simple function to get amount of items currently in the indexing queue.
   *
   * @return int
   *    The mount of items in the queue.
   */
  public function getQueueLength() {

    // Get the queue.
    $queue = $this->queueFactory->get("mongodb_indexer_items");

    if ($queue) {
      $number = $queue->numberOfItems();

      return $number;
    }
  }


  /**
   * Get list of entities without bundle support.
   *
   * There is entity like comment that does not support bundle conditions.
   * And also Those that are ussually don't have many bundles.
   *
   * @return array
   *   list of entity to be excluded from bundle index selector.
   */
  public function getEntityWithoutBundleSupport() {
    return array('comment', 'taxonomy_vocabulary', 'file', 'user');
  }


  /**
   * Get a MongoDB collection object.
   *
   * @param string $collection_name
   *   Name of the collection.
   *
   * @return object
   *   Mongodb Collection object. False if can not connect.
   */
  private function getCollection($collection_name) {
    if (!isset($this->collection[$collection_name])) {
      try {

        if (empty($this->mongoClient)) {
          throw new Exception('Unable to connect to MongoDB.');
        }

        // Get the db from settings.
        $mongodb_settings = $this->config->get('driver_settings');

        $db = isset($mongodb_settings['db']) ? $mongodb_settings['db'] : '';

        // Select the collection.
        $this->collection[$collection_name] = $this->getMongoClient()->selectCollection($db, $collection_name);
      }
      catch (Exception $e) {
        \Drupal::logger('mongodb_indexer')->error($e->getMessage());
        return FALSE;
      }
    }

    return $this->collection[$collection_name];
  }


  /**
   * Returns the name of the collection for the given entity type and bundle.
   *
   * @param string $entity_type
   *   Entity type.
   * @param string $bundle
   *   Bundle name. Optional.
   *
   * @return string
   *   Collection name of format entity_type.bundle or entity_type
   *   (if bundle not given).
   */
  private function getCollectionName($entity_type, $bundle = FALSE) {
    $collection_name = $entity_type;
    if ($bundle && !in_array($entity_type, $this->getEntityWithoutBundleSupport()) && $entity_type != $bundle) {
      $collection_name .= '.' . $bundle;
    }

    return $collection_name;
  }


  /**
   * Update or insert a document to MongoDB.
   *
   * @param string $collection_name
   *   Name of the collection.
   * @param int $document_id
   *   Mongodb document _id.
   * @param array $document
   *   Array of fields data to be updated/inserted.
   *
   * @return Misc
   *   Result return by the operation. False if fail.
   */
  public function upsertDocument($collection_name, $document_id, array $document) {
    $collection = $this->getCollection($collection_name);

    if (!$collection) {
      return FALSE;
    }

    // Construct options for MongoDB operation.
    $options = array(
      'upsert' => TRUE,
    );

    try {
      $result = $collection->updateOne(
        array('_id' => intval($document_id)),
        array('$set' => $document),
        $options
      );
      return $result;
    }
    catch (Exception $e) {
      $error_message = 'Error indexing item id:' . $document_id . ' of collection: ' . $collection_name . '. The error message was: ' . $e->getMessage();
      \Drupal::logger('mongodb_indexer')->error($error_message);
      return FALSE;
    }
  }


  /**
   * Delete a document from MongoDB.
   *
   * @param string $collection_name
   *   Name of the collection.
   * @param int $document_id
   *   Mongodb document _id.
   *
   * @return Misc
   *   Result return by the operation. False if fail.
   */
  public function deleteDocument($collection_name, $document_id) {
    $collection = $this->getCollection($collection_name);
    if (!$collection) {
      return FALSE;
    }
    try {
      $result = $collection->deleteOne(
        array(
          '_id' => intval($document_id),
        )
      );
      return $result;
    }
    catch (Exception $e) {
      $error_message = 'Error deleting item id: ' . $document_id . ' of collection: ' . $collection_name . '. The error message was: ' . $e->getMessage();
      \Drupal::logger('mongodb_indexer')->error($error_message);
      return FALSE;
    }
  }


  /**
   * Check if an entity is of selected entity bundle to index.
   *
   * @param object $entity
   *   The entity to check.
   * @param string $type
   *   The entity's type.
   *
   * @return bool
   *   True/False on valid/invalid.
   */
  private function isValidToIndex($entity, $type) {

    // Get selected entities
    $selected_entity_bundle = $this->getSelectedEntityBundle();

    // We have nothing selected in configs.. Bail out and abort
    if (empty($selected_entity_bundle)) {
      return FALSE;
    }


    // Extract bundle information from the entity
    $bundle = $entity->bundle();


    // Create associative array of the entity data
    $needle = array(
      'entity_type' => $type,
      'bundle' => $bundle,
    );


    $no_bundle_support_entity = $this->getEntityWithoutBundleSupport();
    if (in_array($type, $no_bundle_support_entity)) {
      $needle['bundle'] = FALSE;
    }

    // Now we start checking.
    if (!in_array($needle, $selected_entity_bundle)) {
      return FALSE;
    }
    else {
      return TRUE;
    }
  }


  /**
   * Get selected entity and bundle to index in a more machine readable format.
   *
   * @return array
   *   List of entity and it's bundle that are valid to send to MongoDB.
   */
  public function getSelectedEntityBundle() {
    $selected_entity_bundle = $this->config->get('mongodb_indexer_selected_entity_bundle');

    if (empty($selected_entity_bundle)) {
      return array();
    }

    // The selected_entity_bundle is saved with key is ENTITY_TYPE*BUNDLE
    // OR just ENTITY_TYPE in some cases
    // Only pick those that are not 0
    $return_array = array();
    foreach ($selected_entity_bundle as $key => $val) {
      if ($val !== 0) {
        $entity_bundle_array = explode("*", $key);
        if (!empty($entity_bundle_array)) {
          $return_array[] = array(
            'entity_type' => array_shift($entity_bundle_array),
            'bundle' => !empty($entity_bundle_array) ? array_shift($entity_bundle_array) : FALSE,
          );
        }
      }
    }

    return $return_array;
  }


  /**
   * Mark an item as need to be indexed.
   *
   * This function add the item to the indexer queue.
   *
   * @param object $entity
   *   The entity to be marked as need to index.
   * @param string $type
   *   The entity's type.
   * @param string $queue_type
   *   The queue type:
   *   IndexerEngineInterface::MONGODB_INDEXER_INDEX_TYPE_UPSERT.
   *   IndexerEngineInterface::MONGODB_INDEXER_INDEX_TYPE_DELETE.
   *
   * @return boll
   *   True/False on success/error
   */
  public function markItemChange($entity, $type, $queue_type) {
    static $shutdown_registered = FALSE;
    static $enqueued = array();

    // Check if this item is valid to be indexed.
    if (!$this->isValidToIndex($entity, $type)) {
      return FALSE;
    }

    // Extract required entity information.
    $entity_id = $entity->id();
    $entity_bundle = $entity->bundle();


    // If either of the variables are not set, abort
    if (!isset($entity_id) || !isset($entity_bundle)) {
      return FALSE;
    }

    // Enqueue items only once per page load.
    if (!empty($enqueued["$type-$entity_id-$queue_type"])) {
      return TRUE;
    }


    // Create new queue item
    $item = new \stdClass();
    $item->entity_type = $type;
    $item->bundle = $entity_bundle;
    $item->entity_id = $entity_id;
    $item->queue_type = $queue_type;

    $queue = $this->queueFactory->get("mongodb_indexer_items");
    $queue->createQueue();

    $result = $enqueued["$type-$entity_id-$queue_type"] = $queue->createItem($item);

    // Register a shutdown function to process
    // changed items after everything else is done.
    $index_immediately = $this->config->get('mongodb_indexer_index_immediately');
    // Register shutdown if index immediately is TRUE
    // and do it once.
    // TODO: fix this
    /*if (!$shutdown_registered && $index_immediately) {
      drupal_register_shutdown_function("mongodb_indexer_index_queued_items");
      $shutdown_registered = TRUE;
    }*/
    print_r($result);
    return $result;

  }


  /**
   * Mark items as need to reindex.
   *
   * This function adds the items to the reindexer queue
   * mongodb_indexer_reindex_items
   *
   * @param array $items
   *   List of items to be reindexed.
   *   Each item must have entity_id and entity_type.
   *
   * @return bool
   *   True/False on success/error
   */
  public function markItemsReindex(array $items) {
    $items_to_reindex = array();
    foreach ($items as $item) {
      if (isset($item['entity_type']) && isset($item['entity_id'])) {
        // Now the item should be legit.
        $items_to_reindex[] = $item;
      }
    }

    // Create a queue.
    $item = new \stdClass();
    $item->items = $items_to_reindex;
    $queue =  $this->queueFactory->get("mongodb_indexer_reindex_items");
    $queue->createQueue();
    $result = $queue->createItem($item);

    return $result;
  }


  /**
   * Index predefined anount of items in the queue mongodb_indexer_items.
   *
   * This function should be called from a cron or shut down function.
   * The amount of items to be processed are based on setting variable
   * "mongodb_indexer_items".
   *
   * @return int
   *   Total items processed.
   */
  public function indexQueuedItems() {
    // Get item from queue => convert to document => then index it.
    // Get queue info.
    $queue = $this->queueFactory->get("mongodb_indexer_items");
    $limit_per_run = $this->config->get('mongodb_indexer_items_per_cron_run');

    $total_processed = 0;

    // Loop and claim items until we've got nothing else todo
    // OR has reached the limit.
    while (($item = $queue->claimItem()) && $total_processed < $limit_per_run) {
      // Get the queue data.
      $data = $item->data;

      // Get some data to be used later.
      // Notes: we use entity type as collection name.
      $entity_type = $data->entity_type;
      $entity_id = $data->entity_id;
      $bundle = $data->bundle;

      $collection_name = $this->getCollectionName($entity_type, $bundle);

      // When an item got deleted all we need to do is sending request
      // to delete it from MongoDB.
      if ($data->queue_type == IndexerEngineInterface::MONGODB_INDEXER_INDEX_TYPE_DELETE) {
        $success = $this->deleteDocument($collection_name, $entity_id);
        if ($success) {
          // Done deleting from MongoDB, remove item form queue.
          $queue->deleteItem($item);
          $total_processed++;
        }
        continue;
      }

      // Continue if queue type is not delete.

      $entity = entity_load($entity_type,$entity_id);

      if (empty($entity)) {
        // Can't load entity, it must have been deleted.
        // Just delete from queue.
        $queue->deleteItem($item);
        continue;
      }

      // Convert to MongoDB document.
      $document = $this->convertEntityToMongoDocument($entity, $entity_type);

      $success = FALSE;
      // Save it according to $data->queue_type.
      if ($document) {
        // Save it to it's own collection with the name is entity type.
        $document = $this->prepareDocumentForUpdate($document);
        $success = $this->upsertDocument($collection_name, $entity_id, $document);
      }

      // Done, remove item form queue if operation was a success.
      if ($success) {
        $queue->deleteItem($item);
        $total_processed++;
      }
    }

    return $total_processed;
  }


  /**
   * Convert an entity to mongo document.
   *
   * @param object $entity
   *   The entity to be converted.
   * @param string $entity_type
   *   The entity's type.
   *
   * @return object
   *   The MongoDB document object.
   *   False if the entity is not valid to be added as index.
   */
  private function convertEntityToMongoDocument($entity, $entity_type) {
    // Check if this item is valid to be indexed.
    if (!$this->isValidToIndex($entity, $entity_type)) {
      // Return false to tell this entity is not valid to index.
      return FALSE;
    }

    // Convert entity to MongoDB document.
    // Notes: only get fields and property.
    $document = new \stdClass();


    // Get id, bundle to use later.
    $entity_id = $entity->id();
    $bundle = $entity->bundle();
    $vid = 0;

    if (!$entity_id) {
      return $document;
    }

    // Construct some initial data.
    $document->_id = intval($entity_id);
    $document->_bundle = $bundle;
    $document->_type = $entity_type;

    if (isset($vid)) {
      $document->revision_id = intval($vid);
    }

    $document->indexed_time = new UTCDateTime(round(microtime(TRUE) * 1000));

    // Get the entity fields and their data.
    $fields_info = $entity->getFieldDefinitions();
    foreach ($fields_info as $field_name => $value) {

      $field_value = $this->castingStorageValue($value->getType(), $entity->get($field_name)->value);

      if (isset($field_value)) {
        $document->$field_name = $field_value;
      }
      else {
        $document->$field_name = NULL;
      }
    }


    $dispatcher = \Drupal::service('event_dispatcher');
    $event = new IndexerEvent();
    $event->setDocument($document);
    $event->setEntity($entity);

    $response = $dispatcher->dispatch('mongodb_indexer.build', $event);

    $document = $response->getDocument();

    return $document;
  }


  /**
   * Index a single batch of items in the queue mongodb_indexer_reindex_items.
   *
   * Should be called from a cron or batch.
   *
   * @return Mixed
   *   Integer The number of items that were successfully reindexed.
   *   Boolean FALSE if no item could be claimed from the queue
   *   (possibly signifying that it is empty).
   */
  public function indexQueuedReindexItems() {
    // Get items from queue => convert to documents => then index them.
    // Get queue info.
    $queue = $this->queueFactory->get("mongodb_indexer_reindex_items");

    // The return value will be FALSE if no item could be claimed
    // from the queue.
    $result = FALSE;

    // Claim an item from the queue.
    // A single queue item can contain multiple entities need to be reindexed.
    $item = $queue->claimItem();
    if ($item) {
      // If the queue item was claimed successfully, the return value should be
      // an integer showing the number of items successfully indexed.
      $result = 0;

      // Get the queue data.
      $data = $item->data;
      // Loop through data to process entities.
      if (!empty($data->items)) {
        foreach ($data->items as $row) {
          // Get some data to be used later.
          $entity_type = $row['entity_type'];
          $entity_id = $row['entity_id'];
          $bundle = $row['bundle'];

          // We use entity_type.bundle as collection name, usually.
          $collection_name = $this->getCollectionName($entity_type, $bundle);

          // Load the entity.
          $entity = entity_load($entity_type, $entity_id);

          $entity = $entity ? $entity : FALSE;

          if (empty($entity)) {
            continue;
          }

          // Convert to MongoDB document.
          $document = $this->convertEntityToMongoDocument($entity, $entity_type);
          // Save it.
          if ($document) {
            $document = $this->prepareDocumentForUpdate($document);
            $update_index_result = $this->upsertDocument($collection_name, $entity_id, $document);
            if ($update_index_result !== FALSE) {
              // This entity was indexed without an error.
              // Increase the number of successfully indexed entities.
              $result++;
            }
          }
        }
      }
      // Everything is done, remove item form queue.
      $queue->deleteItem($item);
    }

    return $result;
  }


  /**
   * Convert mongo document object to array.
   *
   * To be used with MongoCollection::update. Any fields that are not changed
   * can be unset in this function to reduce data sending MongoDB.
   *
   * @param object $document
   *   The document object to prepare.
   *
   * @return array
   *   The MongoDB document array.
   *   False if the entity is not valid to be added as index.
   */
  private function prepareDocumentForUpdate($document) {
    // Remove _id.
    unset($document->_id);
    // Convert to array.
    $document = (array) $document;

    return $document;
  }


  /**
   * Casting field storage value.
   *
   * @param string $type
   *   Type of field.
   * @param string $value
   *   Value of field.
   *
   * @return string
   *   Casted value
   */
  private function castingStorageValue($type, $value) {
    switch ($type) {
      case 'int':
      case 'serial':
      case 'entity_reference':
      case 'integer':
      case 'list_boolean':
      case 'field_collection':
      case 'boolean':
      case 'created':
      case 'changed':
        return is_array($value) ? array_map('intval', $value) : intval($value);

      case 'float':
        return is_array($value) ? array_map('floatval', $value) : floatval($value);

      case NULL:
        return NULL;

      default:
        return $value;
    }
  }
}